#include "atomic_helpers.h"
#include "sbi.h"
#include <stddef.h>
#include <stdint.h>
#include <string.h>

void trap_handler();
void trap_handler_wrapper() __attribute__((naked));

void user_entry();

size_t strlen(const char* str) {
    size_t len = 0;
    while(str[len] != '\0') {
        len++;
    }

    return len;
}

int main() {
    csrw(stvec, trap_handler_wrapper);
    csrw(sepc, user_entry);

    size_t sstatus = csrr(sstatus);
    sstatus |= (1 << 5);
    csrw(sstatus, sstatus);

    // _start in entry.s will do the sret
}

void user_entry(void) {
    sbi_set_timer(1000000);
    while(1) { }
}

int puts(const char* str) {
    sbiret ret;
    if((ret = sbi_debug_console_write(str, strlen(str))).error != 0) {
        return -1;
    }

    if((ret = sbi_debug_console_write("\n", 1)).error != 0) {
        return -1;
    }
    return 0;
}

void trap_handler_wrapper() {
    __asm__ volatile("call trap_handler\n"
                     "sret");
}

void trap_handler(void) {
    size_t sepc = csrr(sepc);
    size_t scause = csrr(scause);
    size_t sstatus = csrr(sstatus);

    if(scause == 8) {
        puts("hello world");

        sepc += 4;
        csrw(sepc, sepc);
        goto exit;
    } else {
        if(((sstatus >> 8) & 1) == 0) {
            sepc += 4;
            puts("exception occured in usermode, skipping past");
            csrw(sepc, sepc);
            goto exit;
        } else {
            puts("exception occured in supervisor mode, panic");

            while(1) {
                __asm__ volatile("wfi");
            }
        }
    }

exit:
    return;
}
